c     -*- mode: FORTRAN -*-
c
c     This file is part of krot,
c     a program for the simulation, assignment and fit of HRLIF spectra.
c
c     Copyright (C) 1994-1998 Arnim Westphal
c     Copyright (C) 1998 Jochen Kpper
c     
c     $Id: diagonalize.F,v 1.2 1998/10/09 12:33:39 arnim Exp $


#include <arni.h>


      subroutine rotdi( a, n, np, d, v, e, diaalg, shorti )
c     call diagonalization and eigenvector/eigenvalue sorting routines for COMPLETE hamiltonian matrix in Wang basis

      implicit none

      integer        diaalg, ierr, n, np, shorti
      real*8         a(np,np), d(np), v(np,np), e(np)


      ARNIROT_LAUNCH ( "Launching rotdi." )

#ifdef UNDEFINED
      if( diaalg .eq. 1 ) then
c        Jacobi transformation matrix diagonalization technique
         call jacobi( a, n, np, d, v )
c        sort the eigenvalues before returning
         call eigsrt( d, v, n, np )
      else if( diaalg .eq. 2 ) then
#endif
         ierr = 0
c        tridiagonal diagonalization technique (two steps)
c        - reduction to tridiagonal form
         call tred2( np, n, a, d, e, v )
c        - complete diagonalization of tridiagonal matrix
         call tql2( np, n, d, e, v, ierr )
         if ( ( ierr .ne. 0 ) .and. ( shorti .eq. 0 ) )
     *        write(*,*) "ERROR: diagonalization error for J =", (n - 1)/2, ", l = ", ierr
#ifdef UNDEFINED
      end if
#endif

      return
      end


c-----------------------------------------------------------------------------
      subroutine rotdip( J, iact, dmham, dmfwng, a, aem, aom, aep, aop,
     *                   d, dem, dom, dep, dop, v, vem, vom, vep, vop, er,
     *                   diaalg, shorti )
c     call diagonalization and eigenvector/eigenvalue sorting routines for FACTORIZED Wang hamiltonian SUBMATRICES

      implicit none

c     flag for diagonalization algorithm
      integer        diaalg
c     maximum dimension of unfactorized wang matrix  : dmham  = 2*Jmax+1
      integer        dmham
c     maximum dimension of factorized submatrices E,O: dmfwng = Jmax/2+1
      integer        dmfwng
c     actual size of J block (2J+1)
      integer        iact
c
      integer        ierr
c
      integer        ii, ij, ic, ir
c     current rotational quantum number
      integer        J
c     J modulo 2
      integer        Jmod2
c     actual dimensions of submatrices E-, E+, O-, O+
      integer        nem, nep, nom, nop
c     IO format flag
      integer        shorti

c     original Wang matrix
      real*8         a(dmham,dmham)
c     submatrices E-, E+, O-, O+
      real*8         aem(dmfwng,dmfwng), aep(dmfwng,dmfwng)
      real*8         aom(dmfwng,dmfwng), aop(dmfwng,dmfwng)
c     eigenvalues of Wang (hamiltonian) matrix
      real*8         d(dmham)
c     eigenvalues of submatrices E-, E+, O-, O+
      real*8         dem(dmfwng), dep(dmfwng), dom(dmfwng), dop(dmfwng)
c     vector space for temporary eigenvalue storage
      real*8         er(dmfwng)
c     eigenvectors of Wang matrix
      real*8         v(dmham,dmham)
c     eigenvectors of submatrices E-, E+, O-, O+
      real*8         vem(dmfwng,dmfwng), vep(dmfwng,dmfwng)
      real*8         vom(dmfwng,dmfwng), vop(dmfwng,dmfwng)


      ARNIROT_LAUNCH ( "Launching rotdip." )

      Jmod2 = mod(J,2)
      nem   = J/2
      nep   = nem + 1
      nom   = nem + Jmod2
      nop   = nom

c     factorize Wang hamiltonian matrix a into four submatrices E-, O-, E+ and O+
c     (aem, aom, aep and aop)
      do ij = 1, nem, 1
         ic = 2*ij - 1 + Jmod2
         do ii = 1, nem, 1
            ir = 2*ii - 1 + Jmod2
            aem(ii,ij) = a(ir,ic)
         end do
      end do
      do ij = 1, nom, 1
         ic = 2*ij - Jmod2
         do ii = 1, nom, 1
            ir = 2*ii - Jmod2
            aom(ii,ij) = a(ir,ic)
         end do
      end do
      do ij = 1, nep, 1
         ic = 2*ij - 1 + J
         do ii = 1, nep, 1
            ir = 2*ii - 1 + J
            aep(ii,ij) = a(ir,ic)
         end do
      end do
      do ij = 1, nop, 1
         ic = 2*ij + J
         do ii = 1, nop, 1
            ir = 2*ii + J
            aop(ii,ij) = a(ir,ic)
         end do
      end do

#ifdef UNDEFINED
      if (diaalg.eq.1) then
c        Jacobi transformation matrix diagonalization technique
         call jacobi(aem, nem, dmfwng, dem, vem)
         call jacobi(aom, nom, dmfwng, dom, vom)
         call jacobi(aep, nep, dmfwng, dep, vep)
         call jacobi(aop, nop, dmfwng, dop, vop)
c        sort the eigenvalues before returning
         call eigsrt(dem, vem, nem, dmfwng)
         call eigsrt(dom, vom, nom, dmfwng)
         call eigsrt(dep, vep, nep, dmfwng)
         call eigsrt(dop, vop, nop, dmfwng)
      else if (diaalg.eq.2) then
#endif
         ierr = 0
c        tridiagonal diagonalization technique
         call tred2(dmfwng, nem, aem, dem, er, vem)
         call tql2(dmfwng, nem, dem, er, vem, ierr)
         call tred2(dmfwng, nom, aom, dom, er, vom)
         call tql2(dmfwng, nom, dom, er, vom, ierr)
         call tred2(dmfwng, nep, aep, dep, er, vep)
         call tql2(dmfwng, nep, dep, er, vep, ierr)
         call tred2(dmfwng, nop, aop, dop, er, vop)
         call tql2(dmfwng, nop, dop, er, vop, ierr)
         if( ( ierr .ne. 0 ) .and. ( shorti .eq. 0 ) )
     *        write(*,*) "ERROR: diagonalization error for J =", J, ", l = ", ierr
#ifdef UNDEFINED
      end if
#endif

c     fill d vec and v array in CORRECT ASCENDING order (E+ O- O+ E- E+ ...)
      do ij = 1, nem, 1
         ic = 4*ij
         d(ic) = dem(ij)
         do ii = 1, nem, 1
            ir = 2*ii - 1 + Jmod2
            v(ir,ic) = vem(ii,ij)
         end do
      end do
      do ij = 1, nom, 1
         ic = 4*ij - 2
         d(ic) = dom(ij)
         do ii = 1, nom, 1
            ir = 2*ii - Jmod2
            v(ir,ic) = vom(ii,ij)
         end do
      end do
      do ij = 1, nep, 1
         ic = 4*ij - 3
         d(ic) = dep(ij)
         do ii = 1, nep, 1
            ir = 2*ii - 1 + J
            v(ir,ic) = vep(ii,ij)
         end do
      end do
      do ij = 1, nop, 1
         ic = 4*ij - 1
         d(ic) = dop(ij)
         do ii = 1, nop, 1
            ir = 2*ii + J
            v(ir,ic) = vop(ii,ij)
         end do
      end do

      return
      end


c------------------------------------------------------------------------------
      subroutine tred2( nm, n, a, d, e, z )
c     convert a nondiagonal matrix to tridiagonal form

c     Househoulder reduction of a real symmetric, N by N matrix A, stored in
c     a NM by NM physical array. On output, Z is the orthogonal matrix Q
c     effecting the transformation. D returns the diagonal elements of the
c     tridiagonal matrix, and E the off-diagonal elements, with E(1)=0.
c     Array A is left unchanged.


      implicit none

      integer        i, j, k, l, n, nm
      real*8         a(nm,nm), d(nm), e(nm), z(nm,nm)
      real*8         f, g, h, hh, scale

      ARNIROT_LAUNCH ( "Launching tred2." )

c     copy diagonal and subdiagonal elements of A into Z in order not to destroy array A
      do j = 1, n
         do i = j, n
            z( i, j ) = a( i , j )
         end do
      end do

      do i = n, 2, -1
         l = i - 1
         h = 0.d0
         scale = 0.d0
         if (l.gt.1) then
            do k = 1, l, 1
               scale = scale + dabs(z(i,k))
            end do
            if (scale.eq.0.) then
c              skip transformation
               e(i) = z(i,l)
            else
               do k = 1, l, 1
c                 use scaled a's for transformation
                  z( i, k ) = z( i, k ) / scale
c                 form sigma in H
                  h = h + z( i, k ) * z( i, k )
               end do
               f = z(i,l)
               g = -dsign(dsqrt(h),f)
               e(i) = scale*g
c              now H is equation (11.2.4)
               h = h - f*g
c              store u in the ith row of Z
               z(i,l) = f - g
               f = 0.d0
               do j = 1, l, 1
c                 store u/H in the ith column of Z
                  z(j,i) = z(i,j)/h
c                 form an element of Z*u in G
                  g = 0.d0
                  do k = 1, j, 1
                     g = g + z(j,k)*z(i,k)
                  end do
                  if (l.gt.j) then
                     do k = j+1, l, 1
                        g = g + z(k,j)*z(i,k)
                     end do
                  end if
c                 form an element of p in temporarily unused element of E
                  e(j) = g/h
                  f = f + e(j)*z(i,j)
               end do
c              form K, equation (11.2.11)
               hh = f/(h + h)
c              form q and store in E overwriting p
c              note that E(L)=E(I-1) survives
               do j = 1, l, 1
                  f = z(i,j)
                  g = e(j) - hh*f
                  e(j) = g
c                 reduce Z, equation (11.2.13)
                  do k = 1, j, 1
                     z(j,k) = z(j,k) - f*e(k) - g*z(i,k)
                  end do
               end do
            end if
         else
            e(i) = z(i,l)
         end if
         d(i) = h
      end do

      d(1) = 0.d0
      e(1) = 0.d0
c     begin accumulation of transformation matrices
      do i = 1, n, 1
         l = i - 1
c        this block skipped when I=1
         if ( d(i) .ne. 0. ) then
            do j = 1, l, 1
               g = 0.d0
c              use u and u/H stored in Z to form P*Q
               do k = 1, l
                  g = g + z( i, k ) * z( k, j )
               end do
               do k = 1, l
                  z( k, j ) = z( k, j ) - g * z( k, i )
               end do
            end do
         end if
         d(i) = z(i,i)
c        reset row and column of Z to identity matrix for next iteration
         z(i,i) = 1.d0
         if( l .ge. 1 ) then
            do j = 1, l, 1
               z( i, j ) = 0.d0
               z( j, i ) = 0.d0
            end do
         end if
      end do

      return
      end


c------------------------------------------------------------------------------
      subroutine tql2( dim, n, d, e, z, error )
c     Diagonalize a tridiagonal matrix by the QL algorithm.
c     See Numerical Recipes, chapter 11.3 for a description of the algorithm.

c     Input:  dim       Physical dimension of the matrix.
c     Input:  n         Logical dimension of the matrix.
c     InOut:  d         On input, its first n elements are the diagonal elements
c                       of the tridiagonal matrix.
c                       On output, it returns the eigenvalues.
c     InOut:  e         On input the subdiagonal elements of the tridiagonal matrix
c                       with e(1) arbitrary.
c                       CAUTION: Destroyed on output.
c     InOut:  z         On input, the transformation matrix to obtain the tridiagonal
c                       matrix as obtained from tred2.
c                       If you want to diagonilze a naturally tridiagonal matrix
c                       without using tred2, give the unity matrix.
c                       On output, the i-th column of z returns the normalized
c                       eigenvalue corresponding to d( k ).

c     This implementation of the QL algorithm uses implicit shifts to determine the
c     eigenvalues and eigenvectors of a real, symmetric, tridiagonal matrix. ( Thus
c     all eigenvalues will be real. )
c     Usually input will be a matrix previously reduced to the approbiate form
c     by tred2 (v.s.).


      implicit none

      integer        dim, error, n
      real*8         d( dim ), e( dim ), z( dim, dim )

      integer        i, ii, iter, j, k, l, l1, m
      real*8         b, c, f, g, h, p, r, s, machep
c     machep is a machine dependent parameter specifying the relative precision of floating point arithmetic
      parameter      (machep = 1.0d-14)

      ARNIROT_LAUNCH ( "Launching tql2." )

c     error = 0
      if( n .gt. 1 ) then
c        convenient to renumber the elements of E
         do i = 2, n
            e( i - 1 ) = e( i )
         end do
         e( n ) = 0.d0
         b = 0.d0
         f = 0.d0
         do l = 1, n
            iter = 0
            h = machep * ( dabs( d( l ) ) + dabs( e( l ) ) )
            if( b .lt. h )
     *           b = h
c           look for a single small subdiagonal element to split the matrix
            do m = l, n
               if( dabs( e( m ) ) .le. b )
     *              goto 120
c              e(n) is always zero, so there is no exit through the bottom of the loop
            end do
  120       if( m .ne. l ) then
  130          if( iter .eq. 30 )
     *              goto 1000
               iter = iter + 1
c              form shift
               l1 = l + 1
               g = d( l )
               p = ( d( l1 ) - g ) / ( 2.d0 * e( l ) )
               r = dsqrt( p*p + 1.d0 )
               d( l ) = e( l ) / ( p + dsign( r, p ) )
               h = g - d( l )
               do i = l1, n
                  d( i ) = d( i ) - h
               end do
               f = f + h
c              QL transformation
               p = d( m )
               c = 1.d0
               s = 0.d0
               do i = m-1, l, -1
                  g = c * e( i )
                  h = c * p
                  if( dabs( p ) .ge. dabs( e( i ) ) ) then
                     c = e( i ) / p
                     r = dsqrt( c*c + 1.d0 )
                     e( i+1 ) = s * p * r
                     s = c / r
                     c = 1.d0 / r
                  else
                     c = p / e( i )
                     r = dsqrt( c*c + 1.d0 )
                     e( i + 1 ) = s * e( i ) * r
                     s = 1.d0 / r
                     c = c * s
                  end if
                  p = c * d( i ) - s * g
                  d( i + 1 ) = h + s * ( c * g + s * d( i ) )
c                 form eigenvector
                  do k = 1, n
                     h = z( k, i+1 )
                     z( k, i + 1 ) = s *z( k, i ) + c * h
                     z( k, i )     = c *z( k, i ) - s * h
                  end do
               end do
               e( l ) = s * p
               d( l ) = c * p
               if( dabs( e( l ) ) .gt. b )
     *              goto 130
            end if
            d( l ) = d( l ) + f
         end do

c        order eigenvalues and eigenvectors due to the subroutine eigsrt
         do ii = 2, n
            i = ii - 1
            k = i
            p = d( i )
            do j = ii, n
               if( d( j ) .le. p ) then
                  k = j
                  p = d( j )
               end if
            end do
            if( k .ne. i ) then
               d( k ) = d( i )
               d( i ) = p
               do j = 1, n
                  p = z( j, i )
                  z( j, i ) = z( j, k )
                  z( j, k ) = p
               end do
            end if
         end do
      end if
      goto 1001

c     set error: no convergence to an eigenvalue after 30 iterations
c1000 error = l
 1000 error = error + l

 1001 return
      end


c-----------------------------------------------------------------------------
      subroutine diagc( a, b, c, ener, enei, eneb, vecr, veci, vecb, J, dmham, eb, diaalg, shorti )
c     set up and diagonalize the full complex hamiltonian matrix

      implicit none

      integer        dmham
      integer        diaalg
      integer        i, ii, k, icol, ioff
      integer        J
      integer        shorti
      real*8         a(dmham,dmham), b(dmham,dmham)
      real*8         c(2*dmham,2*dmham)
      real*8         ener(dmham), enei(dmham)
      real*8         eneb(2*dmham), eb(2*dmham)
      real*8         vecr(dmham,dmham), veci(dmham,dmham)
      real*8         vecb(2*dmham,2*dmham)

      ARNIROT_LAUNCH ( "Launching diagc." )

c     prepare matrix in block diagonal form with real and imaginary blocks
      call matpre( a, b, c, 2*J+1, dmham )

#ifdef DEBUG_VERBOSE
      write(*,*) 'debugging module: matrix C'
      do i = 1, 4*J+2, 1
         write(*,'(40(f12.3,2x))') (c(i,k), k = 1, 4*J+2, 1)
      end do
#endif

      call rotdi( c, 4*J+2, 2*dmham, eneb, vecb, eb, diaalg, shorti )

c     copy the combined vector matrix into the correct real and imaginary arrays
c     across the top skipping one column
      icol = 0
      ioff = 2*J+1
      do k = 1, 4*J+2, 2
         icol = icol + 1
         do i = 1, 2*J+1, 1
c           load in the real components
            vecr(i,icol) = vecb(i,k)
c           load in the imaginary components
            veci(i,icol) = vecb(i+ioff,k)
         end do
      end do

c     copy the eigenvalues from doubly degenerate eneb to ener
      i = 1
      do ii = 2, 4*J+2, 2
         ener(i) = eneb(ii)
         i = i + 1
      end do

      return
      end

c------------------------------------------------------------------------------
c     take an N by N real matrix A and an N by N imaginary matrix B, both stored
c     in NMAX by NMAX arrays, and PREpare a 2*N by 2*N MATrix C of the form
c          / A  -B \
c     C = |         |
c          \ B   A /

      subroutine matpre(a, b, c, n, nmax)

      implicit none

      integer        i, k, n, nmax
      real*8         a(nmax,nmax), b(nmax,nmax), c(2*nmax,2*nmax)

      ARNIROT_LAUNCH ("Launching matpre.")

      do k = 1, n, 1
         do i = 1, n, 1
            c(i  ,k  ) =  a(i,k)
            c(i  ,k+n) = -b(i,k)
            c(i+n,k  ) =  b(i,k)
            c(i+n,k+n) =  a(i,k)
         end do
      end do

      return
      end


#ifdef UNDEFINED
c-----------------------------------------------------------------------------
      subroutine jacobi( a, n, np, d, v )
c     fortran routine from NUMERICAL RECIPES which employs the Jacobi
c     transformation method, including corrections for double precision
c     calculation (dabs, dsqrt instead of abs, sqrt)

c     Computes all eigenvalues and eigenvectors of a real symmetric matrix A,
c     which is of size N by N, stored in a physical NP by NP array. On output,
c     elements of A above the diagonal are destroyed. D returns the
c     eigenvalues of A in its first N elements. V is a matrix with the same
c     logical and physical dimensions as A whose columns contain, on output,
c     the normalized eigenvectors of A. NROT is the number of Jacobi
c     rotations which were required.


      implicit none

      integer        Jmax
      parameter      ( Jmax = ARNIROT_JMAX )
c     make arrays large enough for full complex matrices to be worked on
      integer        nmax
      parameter      ( nmax = 4 * Jmax + 2 )

      integer        i, ip, iq, j, n, np, nrot
      real*8         a(np,np), d(np), v(np,np)
      real*8         b(nmax), z(nmax)
      real*8         c, g, h, s, sm, t, tau, theta, tresh

      ARNIROT_LAUNCH ( "Launching jacobi." )

c     initialize to the identity matrix
      do ip = 1, n
         do iq = 1, n
            v( iq, ip ) = 0.d0
         end do
         v( ip, ip ) = 1.d0
      end do

c     initialize B and D to the diagonal of A
      do ip = 1, n
         b( ip ) = a( ip, ip )
         d( ip ) = a( ip, ip )
c        vector z will accumulate terms of the form ta_pq as in equation (11.1.14)
         z( ip ) = 0.d0
      end do

      nrot = 0
      do i = 1, 50, 1
         sm = 0.d0
c        sum off-diagonal elements
         do ip = 1, n-1
            do iq = ip+1, n
               sm = sm + dabs( a( ip, iq ) )
            end do
         end do
c        normal return, which relies on quadratic convergence to machine underflow
         if( sm.eq.0.d0 ) goto 999
c        ... on the first three sweeps
         if( i .lt. 4 ) then
            tresh = 0.2d0 * sm / n**2
c           ... thereafter
         else
            tresh = 0.d0
         end if

         do ip = 1, n-1
            do iq = ip+1, n
               g = 100.d0 * dabs( a( ip, iq ) )
c              after four sweeps, skip the rotation if the off-diagonal element is small
               if ((i.gt.4) .and. (dabs(d(ip))+g .eq. dabs(d(ip))) .and. (dabs(d(iq))+g .eq. dabs(d(iq)))) then
                  a(ip,iq) = 0.d0
               else if (dabs(a(ip,iq)) .gt. tresh) then
                  h = d(iq) - d(ip)
                  if (dabs(h)+g .eq. dabs(h)) then
                     t = a(ip,iq)/h
                  else
                     theta = 0.5*h/a(ip,iq)
                     t = 1.d0/(dabs(theta) + dsqrt(1.d0 + theta**2))
                     if (theta.lt.0.) t = -t
                  end if
                  c = 1.d0/dsqrt(1.d0 + t**2)
                  s = t*c
                  tau = s/(1.d0 + c)
                  h = t*a(ip,iq)
                  z(ip) = z(ip) - h
                  z(iq) = z(iq) + h
                  d(ip) = d(ip) - h
                  d(iq) = d(iq) + h
                  a(ip,iq) = 0.d0
c                 case of rotations 1 <= j < p
                  do j = 1, ip-1, 1
                     g = a(j,ip)
                     h = a(j,iq)
                     a(j,ip) = g - s*(h + g*tau)
                     a(j,iq) = h + s*(g - h*tau)
                  end do
c                 case of rotations p < j < q
                  do j = ip+1, iq-1, 1
                     g = a(ip,j)
                     h = a(j,iq)
                     a(ip,j) = g - s*(h + g*tau)
                     a(j,iq) = h + s*(g - h*tau)
                  end do
c                 case of rotations q < j <= n
                  do j = iq+1, n, 1
                     g = a(ip,j)
                     h = a(iq,j)
                     a(ip,j) = g - s*(h + g*tau)
                     a(iq,j) = h + s*(g - h*tau)
                  end do
                  do j = 1, n, 1
                     g = v(j,ip)
                     h = v(j,iq)
                     v(j,ip) = g - s*(h + g*tau)
                     v(j,iq) = h + s*(g - h*tau)
                  end do
                  nrot = nrot + 1
               end if
            end do
         end do

         do ip = 1, n
            b( ip ) = b( ip ) + z( ip )
c           update D with the sum of ta_pq and reinitialize Z
            d( ip ) = b( ip )
            z( ip ) = 0.d0
         end do

      end do
      pause '50 iterations should never happen'

  999 continue
      ARNIROT_DEBUG1( "number of Jacobi rotations :", nrot )

      return
      end


c------------------------------------------------------------------------------
      subroutine eigsrt(d, v, n, np)
c     sort the eigenvalues and eigenvectors

c     Given the eigenvalues D and eigenvectors V as output from JACOBI
c     this routine sorts the eigenvalues into ASCENDING order (not
c     descending, as printed in former editions of NUMERICAL RECIPES), and
c     rearranges the columns of V correspondingly. The method is straight
c     insertion.


      implicit none

      integer        i, j, k, n, np
      real*8         d(np), v(np,np), p

      ARNIROT_LAUNCH ( "Launching eigsrt." )

      do i = 1, n-1
         k = i
         p = d(i)
         do j = i+1, n
c           correction: .ge. changed to .le. to sort into ASCENDING order
            if( d(j) .le. p ) then
               k = j
               p = d(j)
            end if
         end do
         if( k .ne. i ) then
            d(k) = d(i)
            d(i) = p
            do j = 1, n
               p = v(j,i)
               v(j,i) = v(j,k)
               v(j,k) = p
            end do
         end if
      end do
      
      return
      end
#endif
